/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * @(#) $Header: /home/barad-dur/b/users/elan/src/vgw/vgw/ipad/RCS/encoder-ipadvq.h,v 1.6 1995/11/22 03:28:25 elan Exp $ (UC Berkeley) 
 */

#ifndef vgw_ipadvq_h
#define vgw_ipadvq_h

#include "encoder.h"

class IpadChannel;

/* 
 * RescaleInfo is all of the aux info needed to do fast rescaling 
 * of an image from the original mpeg sized image to a reduced VQ 
 * image.  There is one per transformation - i.e. one for the L-Y 
 * rescaling and one for the CrCb->IQ transformations. 
 */
class RescaleInfo {
public:
	~RescaleInfo();

	/* orig_width, orig_height, scaled_width, scaled_height are the
	 * dimensions of the original and scaled images 
	 */
	int     orig_width;
	int     orig_height;
	int     scaled_width;
	int     scaled_height;

	/* orig_bytes_per_line and scaled_bytes_per_line allow selecting
	 * parts of the source and scaled image for zooming and padding 
	 */
	int     orig_bytes_per_line;
	int     scaled_bytes_per_line;

	/* x_orig2scaledlinebuf is used if the original width is larger than
	 * the scaled width.  It maps from the X coord in the original image
	 * into the element in linebuf which corresponds to the scaled image
	 * position.  It will be used to select into which linebuf bin each
	 * original pixel goes 
	 */
	int   **x_orig2scaledlinebuf;

	/* x_scaled2origoffset is used if the original width is smaller than
	 * or the same as the scaled width.  It maps from the scaled X coord
	 * to the original's x coord.  It will be used to select from which
	 * original pixel each linebuf pixel comes 
	 */
	int    *x_scaled2origoffset;

	/* linebuf is where the X size-scaled array is kept.  It is the sum
	 * of all of the pixels falling into it from all of the unscaled rows
	 * corresponding to the current scaled pixel. If the original height
	 * is smaller than this just collects for one original row. 
	 */
	int    *linebuf;

	/* x_scaled2val_scale_large and x_scaled2val_scale_small are used to
	 * map from the shifted line_buf value which range from 0 to 511 into
	 * pixel values ranging from 0 to 255. For each scaled bin they point
	 * to one of the four xval_scale arrays.  x_scaled2val_scale_large is
	 * used if the original height is smaller or equal to the scaled
	 * height or if the original height is larger but this scaled y
	 * corresponds to the larger of the y quantization steps.
	 * x_scaled2val_scale_small is used if the original height is larger
	 * and this scaled y corresponds to the smalled of the y quantization
	 * steps.  
	 */
	u_char **x_scaled2val_scale_large;
	u_char **x_scaled2val_scale_small;
	
	/* orig_rows_per_scaled_y counts how many source rows correspond to
	 * the current scaled row.  If the source height is larger then this
	 * will determine how many rows to collect.  If the source height is
	 * smaller then it will say 1 to get new values or 0 to skip.  In
	 * this case the first entry will always be 1. The sum of the values
	 * will always be the source image height.  
	 */
	int    *orig_rows_per_scaled_y;

	/* The value for the smaller rows just for quickly determining
	 * whether to use small or large arrays 
	 */
	int     small_rows_per_scaled_y;

	/* xval_scale holds the scale values pointed to by
	 * x_scaled2val_scale_large and x_scaled2val_scale_small There are 4
	 * arrays of 512 values 
	 */
	u_char  xval_scale[512 * 4];

	/* shift_bits_large_y and shift_bits_small_y are the number of bits
	 * to shift to get the value into 0..511 range before indexing into
	 * the scale array.  They are indexed by scaled X column 
	 */
	int    *shift_bits_large_y;
	int    *shift_bits_small_y;
};

class IpadVQEncoder : public Encoder {
 public:
	IpadVQEncoder(IpadChannel*, int, int, int, int);
	int encode(u_char*, int, int, u_char*);
 private:
	void InitVQ(int, int);
	void init_CrCb_to_IQ(void);
	int log2_int(unsigned int);
	void ffree(void *);
	RescaleInfo* init_rescale_info(int, int, int, int, int, int);
	void rescale_image(u_char*, u_char*, RescaleInfo*, int, int, int, int);
	void convert_CrCb_IQ(u_char*, u_char*, int, int, int);
	void adapt_codebook(void);
	void adapt_compute_code_sums(u_char*, u_char*, int*, int*, int, int, 
				     int);
	void recompute_codebook(u_char*, int*, int*);
	u_char BestCode(u_char*, u_char*, int, u_char, int*);
	u_char BestCodeHalfRes(u_char*, u_char*, int, u_char);
        u_char BestCodeUniform(u_char*, u_char*, int);
	u_char BestCodeHalfResUniform(u_char*, u_char*, int);
	int BestCodeFast(u_char*, u_char*, int);
 	int BestCodeHalfResFast(u_char*, u_char*, int);
        int vector_quantize_block(u_char*, u_char*, u_char*, int, int, int);
	void dump_vectors(char*, u_char*, int, int, int);
	void compute_block_errors(u_char*, u_char*, int*, u_char*, int, int, 
				  int);
	void prune_codebook(u_char*, int*, int*, int, int, char);
	void supplement_codebook(u_char*, u_char*, int*, int*, int*, int,
				 int, int, char);
 	void generate_fast_codebooks(u_char*, u_char*, u_char*);
#define CODE_VECTORS    256
#define VECTOR_BYTES    16
	/* 
	 * These three are the scaled L,Cr,Cb image which is converted in 
	 * place to Y,I,Q.  They are for the current frame or if
	 * adaptation is occuring then they store all of the adaptation frames
	 * between resizing and recalculation 
	 */
	u_char *L_Y_scaled_buf;
	u_char *Cr_I_scaled_buf;
	u_char *Cb_Q_scaled_buf;
	/* This one is for one frame of interleaving I and Q */
	u_char *IQ_temp_codes;
	/* 
	 * The next three are VQ codes for the frame current frame or if 
	 * adaptation is occuring then they store all of the adaptation 
	 * frames between resizing and recalculation 
	 */
	u_char *Y_frame_data_codes;
	u_char *I_frame_data_codes;
	u_char *Q_frame_data_codes;
	/* 
	 * The next three are the amount of error that each 4x4 block had when 
	 * matching.  Used only for adaptation - they store all of the 
	 * adaptation frames
	 */
	int *Y_frame_coding_errors;
	int *I_frame_coding_errors;
	int *Q_frame_coding_errors;

	/* The current codebook */
	u_char y_codes[CODE_VECTORS * VECTOR_BYTES];
	u_char i_codes[CODE_VECTORS * VECTOR_BYTES];
	u_char q_codes[CODE_VECTORS * VECTOR_BYTES];

	/* Sums for k-means adaptation */
	int y_code_sums[CODE_VECTORS * VECTOR_BYTES];
	int i_code_sums[CODE_VECTORS * VECTOR_BYTES];
	int q_code_sums[CODE_VECTORS * VECTOR_BYTES];

	/* Counts for k-means adaptation */
	int y_code_counts[CODE_VECTORS];
	int i_code_counts[CODE_VECTORS];
	int q_code_counts[CODE_VECTORS];

	/* Maximum y code self-similarity for adaptation */
	int y_codes_sim_max[CODE_VECTORS][CODE_VECTORS];
	int i_codes_sim_max[CODE_VECTORS][CODE_VECTORS];
	int q_codes_sim_max[CODE_VECTORS][CODE_VECTORS];

	/* Average y code self-similarity for adaptation */
	int y_codes_sim_ave[CODE_VECTORS][CODE_VECTORS];
	int i_codes_sim_ave[CODE_VECTORS][CODE_VECTORS];
	int q_codes_sim_ave[CODE_VECTORS][CODE_VECTORS];

	/* 
	 * For use with error_tolerance - reset at beginning and codebook 
	 * change 
	 */
	int prev_codes_valid;
	/* 
	 * Number of frames stored for adaptation - 0 if adaptation not going 
	 * on.
	 */
	int adapt_frames_stored;

	/* Number of frames more which should be stored for adaptation */
	int adapt_frames_needed;
	int _squares[512];
	int* squares;
	unsigned short CrCb_to_IQ[256 * 256];
#ifdef STATISTICS
	unsigned long low_error;
	unsigned long max_error;
	unsigned long total_error;
	unsigned long total_vectors;
#endif
	unsigned int inWidth, inHeight;
	unsigned int srcX;
	unsigned int srcY;

	unsigned int srcWidth; /* If srcWidth or srcHeight don't get set */
	unsigned int srcHeight;/* in the command line switches then they */
	unsigned int outWidth; /* default to inWidth and inHeight in InitVQ. */
	unsigned int outHeight; /* Default output width to match Infopad. */

	/* Default output height to match Infopad. */
	unsigned int outSize;		/* outWidth * outHeight */
	int     codeHeight;	/* Number of Y codes high the picture is */
	int     codeWidth;	/* Number of Y codes wide the picture is */
	int     codeSize;	/* Number of Y codes total the picture is */
	int skipped_frames;
	int     dump_codebook;
	int     halfres_coding;
	int     uniform_coding;
	int     fast_coding;
	char   *codebook_file;
	int     no_VQconv;
	float   frames_per_sec_override;
	int     adapt_frames;
	int     error_tolerance;
	int     print_vq_error;
	int     adapt_global_threshold;
	int     skip_every;

	/* From vq_code.cc */
	u_char *y_codes_addr;
	u_char *i_codes_addr;
	u_char *q_codes_addr;
	long    cb_squares[512];
	u_char  solid_2_code[3][256];
	u_char  diag0_2_code[3][64][64];
	u_char  diag1_2_code[3][64][64];
	u_char  horiz_2_code[3][64][64];
	u_char  vert_2_code[3][64][64];

	IpadChannel* net_;
	int dump_;
	int decimation_;
	RescaleInfo *L_rescale_info, *CrCb_rescale_info;
 public:
	inline IpadChannel* net(void) { return net_; }
};

#endif /* vgw_ipadvq_h */






