/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * @(#) $Header: /home/barad-dur/b/users/elan/src/vgw/vgw/ipad/RCS/ipadchan.cc,v 1.3 1995/11/22 03:28:25 elan Exp $ (UC Berkeley)
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <errno.h>
#include "ipadchan.h"

#include "global.h"
#include "packets.h"
#include "ipnio.h"
#include "ipnproto.h"

#define MAX_VIDEO_PACKET_LONGS 512

#ifndef min
#define min(a, b) ((a) > (b))? (b) : (a)
#endif

IpadChannel::IpadChannel(int p)
	    :Network(N_IPAD), pad_id_(p)
{
#ifdef notdef
	fprintf(stderr,"Sending VQ directly to pad %d\n", p);
#endif
	if ((ipad_conn_ = (void *)VideoConnectToPS(p)) == 0) {
		char buf[100];
		sprintf(buf, 
			"Can't connect to pad server for pad %d, error %d",
			p, errno);
		perror(buf);
		exit(-1);
	}
}

/*
 *--------------------------------------------------------------
 *
 * send_codebook_entries
 *
 *      Sends codebook update message over InfoPad network.
 *
 *	Input:
 *		which_codebook:		0=Y, 1=I, 2=Q
 *		first_entry:		First entry to be updated
 *		num_entries:		Number of entries to be updated
 *--------------------------------------------------------------
 */
void 
IpadChannel::send_codebook_entries(u_int which_codebook,
					   u_int first_entry, 
					   u_int num_entries,
					   u_char *codebook_data)
{
	int i,j;
	u_char A, B, C, D;
	u_char* xlat;		/* Points into Y or IQ xlat */
	u_long	message[256];	/* Update data message */
	u_char* in_ptr;
	u_long* out_ptr;
	int start_addr;

	/* 
	 * y_xlat and iq_xlat do bit translations for hardware
	 * y_xlat does just a simple shifting 
         * iq_xlat does offset binary->sign mag w/ flip & shift
	 */
	static int xlats_initted=0;
	static u_char y_xlat[256];	
	static u_char iq_xlat[256];
	if (!xlats_initted) {
		for (j=0; j<128; j++) {
			y_xlat[j] = j >> 2;
			y_xlat[j + 0x80] = (j >> 2) | 0x20;
			iq_xlat[j] 	= ((j >> 2)^0x1f) | 0x20;
			iq_xlat[j + 0x80] =  (j >> 2);
		}
	}

	/* 
	 * We have to reformat codebook data to pack it into
	 * 6 bit quantities, switch it from entry-major order to
	 * entry-row major order, and finally switch I and Q into
	 * sign magnitude format
	 * Hardware wants reverse of:
	 * 00000000 DDDDDDCC CCCCBBBB BBAAAAAA
	 * which is:
	 * BBAAAAAA CCCCBBBB DDDDDDCC 00000000
	 * 
	 * In code, t0=AAAAAA, t1=BBBBBB, t2=CCCCCC, t3=DDDDDD   
	 */
	if ((first_entry + num_entries > 256) || (num_entries == 0) || 
	    (which_codebook > 2)) {
		fprintf(stderr, 
			"Bad call to send_codebook_entries(%d, %d, %d, XXX)\n",
			which_codebook, first_entry, num_entries);
		return;
	}

	if (which_codebook == 0)
		xlat = y_xlat;
	else
		xlat = iq_xlat;

	/* 
	 * Reformat data and send.  It is passed (and written in
	 * files as: 
	 *  entry0line0, entry0line1, entry0line2, entry0line3, 
	 *  entry1line0, entry1line1, entry1line2, entry1line3, 
	 *  entry2line0, entry2line1, entry2line2, entry2line3, 
	 * 		.......					
	 *  entry255line0, entry255line1, entry255line2,entry255line3,
	 *  		etc but the hardware wants 
	 *  entry0line0, entry1line0, ..., entry255line0, 
	 *  entry0line1, entry1line1, ..., entry255line1, 
	 *  entry0line2, entry1line2, ..., entry255line2, 
	 *  entry0line3, entry1line3, ..., entry255line3,
	 */ 
	for (j=0; j<4; j++) {
		out_ptr = message;
		in_ptr = codebook_data + j*4;

		for (i=0; i<num_entries; i++) {
			A = xlat[*in_ptr++];
			B = xlat[*in_ptr++];
			C = xlat[*in_ptr++];
			D = xlat[*in_ptr++];
			in_ptr += 12;
			*out_ptr++ = (A << 24) |
				     ((B & 0x3c) << 14) |
				     ((B & 0x03) << 30) |
				     ((C & 0x30) <<  4) |
				     ((C & 0x0f) << 20) |
				     (D << 10);
		}
		start_addr = which_codebook*1024 + (first_entry & 0xff) + 
			j*256;
		send_video_data(T_VIDEOCODE, num_entries, start_addr,
				(u_char *)message, 0);
	}
}


/*
 *--------------------------------------------------------------
 *
 * send_video_data_infopad
 *
 *      Sends video (frame or codebook) messages packetizing when necessary.
 *
 *	Input:
 *		type:			T_VIDEOCODE or T_VIDEOFRAME
 *		num_longs:		Number of longwords to send
 *		start_addr:		Start address of destination
 *		data:			The data to send
 *		flip_on_last_packet:	Should frame buffer be flipped?
 *--------------------------------------------------------------
 */
void 
IpadChannel::send_video_data(int type, int num_longs, int start_addr, 
				     u_char* data, int flip_on_last_packet)
{
	u_long headl;
	int words_left;
	int words_this_xfer;
	int cur_addr;
	u_char* data_ptr;
	static char msg_buf[MAX_VIDEO_PACKET_LONGS*4 + 4];
	int bytesWritten;
	
	static int no_codebook_send_hack_helper=0;

	cur_addr = start_addr;
	data_ptr = data;

	for (words_left=num_longs; words_left>0; words_left-=words_this_xfer) {
		words_this_xfer = min(words_left, MAX_VIDEO_PACKET_LONGS);

		headl = cur_addr;

		if (flip_on_last_packet && (words_left == words_this_xfer))
			headl |= 0x0400;

		msg_buf[0] = headl & 0xff;
		msg_buf[1] = (headl >> 8) & 0xff;
		msg_buf[2] = (headl >> 16) & 0xff;
		msg_buf[3] = (headl >> 24) & 0xff;

		memcpy(msg_buf+4, data_ptr, words_this_xfer*4);
		
		IPNSetType((PORTFILE *)ipad_conn_, (IPNtype)type);
		bytesWritten = IPNWrite(msg_buf, words_this_xfer*4+4, 1, 
					   (PORTFILE *)ipad_conn_);
		if (bytesWritten != words_this_xfer*4+4) {
			fprintf(stderr, "Couldn't send %d bytes to pad server"
				      " (only sent %d bytes)\n",
	      				words_this_xfer*4+4, bytesWritten);
		}

		if (!no_codebook_send_hack_helper) {
			IPNSetType((PORTFILE *)ipad_conn_, (IPNtype)type);
			bytesWritten = IPNWrite(msg_buf, words_this_xfer*4+4, 
						1, (PORTFILE *)ipad_conn_);
			if (bytesWritten != words_this_xfer*4+4) {
				fprintf(stderr, 
					"Couldn't send %d bytes to pad server"
					" (only sent %d bytes)\n",
					words_this_xfer*4+4, bytesWritten);
			}
			no_codebook_send_hack_helper=1;
		}
		data_ptr += words_this_xfer * 4;
		cur_addr += words_this_xfer;
	}
}







