#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include "transcoder.h"
#include "crdef.h"
#include "transmitter.h"

Transcoder::Transcoder(int hdrlen)
	   :PacketHandler(hdrlen), fc_(0.), forward_(0)
{
	*tname_ = '\0';
}


int Transcoder::command(int argc, const char*const* argv)
{
	if (argc < 2) 
		return (TCL_ERROR);

	Tcl& tcl = Tcl::instance();
	const char* cmd = argv[1];
	if (argc == 2) {
		if (strcmp(cmd, "media-type") == 0) {
			tcl.resultf("%s", media_type());
			return (TCL_OK);
		}
		if (strcmp(cmd, "forward") == 0) {
			tcl.resultf("%d", forward_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "tname") == 0) {
			tcl.resultf("%s", tname_);
			return (TCL_OK);
		}
	}
	if (argc == 3) {
		if (strcmp(cmd, "forward") == 0) {
			forward_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(cmd, "tname") == 0) {
			strcpy(tname_, argv[2]);
			return (TCL_OK);
		}
	}
	return (TclObject::command(argc, argv));
}

VideoTranscoder::VideoTranscoder(int hdrlen)
		:Transcoder(hdrlen), encoder_(0), crvec_(0), 
		 inw_(0), inh_(0), outw_(0), outh_(0)
{}

VideoTranscoder::~VideoTranscoder()
{
	delete encoder_;
}

void VideoTranscoder::send_all(void)
{
	if (crvec_ != 0)
		for (int i = 0; i < nblk_; i++)
			crvec_[i] = CR_MOTION|CR_SEND;
}

void VideoTranscoder::crinit(int w, int h)
{
	int blkw = w >> 3;
	int blkh = h >> 3;
	scan_ = 0;
	nblk_ = blkw * blkh;
	delete crvec_;
	crvec_ = new u_char[nblk_];
	send_all();
	rover_ = 0;
}

void VideoTranscoder::age_blocks(void)
{
	for (int i = 0; i < nblk_; i++) {
		int s = CR_STATE(crvec_[i]);
		/*
		 * Age this block.
		 * Once we hit the age threshold, we
		 * set CR_SEND as a hint to send a
		 * higher-quality version of the block.
		 * After this the block will stop aging,
		 * until there is motion.  In the meantime,
		 * we might send it as background fill
		 * using the highest quality.
		 */
		if (s <= CR_AGETHRESH) {
			if (s == CR_AGETHRESH)
				s = CR_IDLE;
			else {
				if (++s == CR_AGETHRESH)
					s |= CR_SEND;
			}
			crvec_[i] = s;
		} else if (s == CR_BG)
			/*
			 * reset the block to IDLE if it was sent
			 * as a BG block in the last frame.
			 */
			crvec_[i] = CR_IDLE;
	}
	/*
	 * Now go through and look for some idle blocks to send
	 * as background fill.
	 */
	int blkno = rover_;
#ifdef notyet
	int n = (delta_ * 2. < frametime_)? idle_high_ : idle_low_;
#endif
	int n = 2;
	while (n > 0) {
		int s = CR_STATE(crvec_[blkno]);
		if (s == CR_IDLE) {
			crvec_[blkno] = CR_SEND|CR_BG;
			--n;
		}
		if (++blkno >= nblk_) {
			blkno = 0;
			/* one way to guarantee loop termination */
			break;
		}
	}
	rover_ = blkno;

	/*
	 * Bump the CR scan pointer.  This variable controls which
	 * scan line of a block we use to make the replenishment
	 * decision.  We skip 3 lines at a time to quickly precess
	 * over the block.  Since 3 and 8 are coprime, we will
	 * sweep out every line.
	 */
	scan_ = (scan_ + 3) & 7;
}


int VideoTranscoder::command(int argc, const char*const* argv)
{
	if (argc < 2) 
		return (TCL_ERROR);
	
	const char* cmd = argv[1];
	if (argc == 3) {
		if (strcmp(cmd, "encoder") == 0) {
			delete encoder_;
			encoder_ = (VideoEncoder*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
	}

	return (Transcoder::command(argc, argv));
}


#define	INITIAL_OFFSET (4*FRAMESIZE)

AudioTranscoder::AudioTranscoder(int hdrlen)
		:Transcoder(hdrlen), encoder_(0),
		 var_(INITIAL_OFFSET << (VAR_FILTER - VAR_MULT)),
		 hostoffset_(0), predicted_drop_(-1), lecture_mode_(0),
		 playout_(INITIAL_OFFSET << PLAYO_FILTER), maxdel_(0), 
		 block_size_(FRAMESIZE), framesize_(FRAMESIZE)/*XXX*/
{
}

AudioTranscoder::~AudioTranscoder()
{
	delete encoder_;
}

int AudioTranscoder::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);

	Tcl& tcl = Tcl::instance();

	const char* cmd = argv[1];
	if (argc == 2) {
		char* bp = tcl.buffer();
		if (strcmp(cmd, "block-size") == 0) {
			sprintf(bp, "%u", block_size_);
			tcl.result(bp);
			return (TCL_OK);
		}
		if (strcmp(cmd, "playout") == 0) {
			sprintf(bp, "%u", playout());
			tcl.result(bp);
			return (TCL_OK);
		}
		if (strcmp(cmd, "reset-offset") == 0) {
			hostoffset_ = 0;
			var_ = 0;
			playout_ = 0;
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(cmd, "lecture-mode") == 0) {
			lecture_mode_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(cmd, "encoder") == 0) {
			delete encoder_;
			encoder_ = (AudioEncoder*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
	}
	return (Transcoder::command(argc, argv));
}

inline int absdiff(int x, int y)
{
	register int r = y - x;
	return (r < 0? -r : r);
}

static inline int newoffset(
	int nvar,
	int playout,
	int maxdel, int mindel, int lecture)
{
	register int offset = nvar;
	if (offset > maxdel)
		offset = maxdel;
	register int diff = playout - offset;

	if (diff > 0) {
		// offset going down: in LectureMode, drop at most
		// one frametime per talkspurt.  In ConferenceMode,
		// drop at most 1/2 of difference.
		if (lecture) {
			if (diff > FRAMESIZE) {
				if (playout > (maxdel * 3) / 4 &&
				    diff > 10 * FRAMESIZE)
					diff = 5 * FRAMESIZE;
				else
					diff = FRAMESIZE;
			}
		} else
			diff >>= 1;
		offset = playout - diff;
	} else if (-diff > maxdel) {
		// offset going way up: only allow 3/4 of max.
		offset = (maxdel * 3) / 4;
	}
	if (offset > (maxdel * 7) / 8)
		offset = (maxdel * 7) / 8;
	else if (offset < mindel)
		offset = mindel;
	return (offset);
}

int AudioTranscoder::adapt(const rtphdr* rh, u_int32_t local_clock)
{
	register u_int32_t tstamp = ntohl(rh->rh_ts);
	register int hoff = (int)hostoffset_;
	register int offset = (tstamp + hoff - local_clock) &~ 3;
	int new_ts = rh->rh_flags & htons(RTP_M);
	if (hoff == 0 || new_ts) {
		/*
		 * start of new talk spurt --
		 * use accumulated variance to compute new offset if
		 * this would make a significant change.  We change if
		 *  - the variance is currently 'small', or
		 *  - the change would be a least a packet time
		 */
		register int nvar = var_ >> (VAR_FILTER - VAR_MULT);
		offset = playout_ >> PLAYO_FILTER;
		if (nvar < 3*FRAMESIZE || absdiff(nvar, offset) >= FRAMESIZE) {
			offset = newoffset(nvar, offset, maxdel_, 
					   block_size_, lecture_mode_);
			/*
			 * assume that a talk spurt starts with TALK_LEAD
			 * samples of history & subtract them off if possible.
			 */
			if (new_ts) {
				offset -= 4 * FRAMESIZE;
				if (offset < block_size_)
					offset = block_size_;
			}
		}
		hostoffset_ = local_clock - tstamp + offset;
	} else if (offset < 0 || offset > maxdel_) {
		/*
		 * packet out of range -- if last packet also out of
		 * range or if the delay would increase, resync.
		 */
		if (offset < 0 || predicted_drop_ == tstamp) {
			offset = newoffset(var_ >> (VAR_FILTER - VAR_MULT),
					   playout_ >> PLAYO_FILTER,
					   maxdel_, block_size_,
					   lecture_mode_);
			hostoffset_ = local_clock - tstamp + offset;
		} else {
			predicted_drop_ = tstamp + block_size_;
			lastrecv_ = tstamp - local_clock;
			return (-1);
		}
	} else {
		// packet in range, update interarrival var. est.
		register int nvar = var_;
		register int off = tstamp - local_clock - lastrecv_;
		if (off < 0)
			off = -off;
		off -= (nvar >> VAR_FILTER);
		var_ = nvar + off;
	}
	lastrecv_ = tstamp - local_clock;
	register u_int avgplay = playout_;
	playout_ = avgplay + (offset - (avgplay >> PLAYO_FILTER));
	offset &= ~3;
	delvar_ = var_ >> VAR_FILTER;
	return (offset);
}

static class NullTranscoderMatcher : public Matcher {
public:
	NullTranscoderMatcher() : Matcher("transcoder") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "null") == 0)
			return (new NullTranscoder);
		return (0);
	}
} null_transcoder_matcher;

int NullTranscoder::recv(const rtphdr* rh, const u_char* bp, int cc, 
			 Transmitter& tx)
{
	if (forward_) {
		/* No rate control - just send the packet through. */
		memcpy(pktbuf_, rh, sizeof(*rh));
		memcpy(pktbuf_ + sizeof(*rh), bp, cc);
		tx.forward(pktbuf_, cc + sizeof(*rh));
		return (0);
	}
	if (tx.bps() == 0)
		return (0);
	/* 
	 * Rate control.
	 * This actually bogus, since we are dropping random packets,
	 * but the alternative is not to let any forwarding to 
	 * occur in rate controlled sessions.  May as well let the user
	 * decide.
	 */
	double now = gettimeofday();
	if (fc_ <= now) {
		if (now - fc_ > 200000.) 
			fc_ = now;
		memcpy(pktbuf_, rh, sizeof(*rh));
		memcpy(pktbuf_ + sizeof(*rh), bp, cc);
		tx.forward(pktbuf_, cc + sizeof(*rh));
		int bits = 8 * (cc + sizeof(*rh));
		fc_ += 1e6 * bits / tx.bps();
	}
	return (0);	/* Don't know frame boundaries. */
}
