/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/video/RCS/transcoder-h261.cc,v 1.4 1996/08/26 21:58:28 elan Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <netinet/in.h>
#include "inet.h"
#include "rtp.h"
#include "transcoder.h"
#include "transmitter.h"
#include "p64/p64.h"
#include "encoder.h"
#include "crdef.h"

class H261Transcoder : public VideoTranscoder {
public:
	H261Transcoder(int dct);
	~H261Transcoder();
	virtual int recv(const rtphdr* rh, const u_char* bp, int cc, 
			 Transmitter& tx);
protected:
	void configure();
	P64Decoder* decoder_;
	int dct_;
	int ibit_;
	int h261_rtp_bug_;
};

static class H261TranscoderMatcher : public Matcher {
public:
	H261TranscoderMatcher() : Matcher("transcoder") {}
	TclObject* match(const char* fmt) {
		    if (strcmp(fmt, "h261/dct") == 0)
			    return (new H261Transcoder(1));
		    if (strcmp(fmt, "h261/pixel") == 0)
			    return (new H261Transcoder(0));
		    return (0);
	}
} h261_transcoder_matcher;

H261Transcoder::H261Transcoder(int dct)
	       : VideoTranscoder(sizeof(h261hdr)), dct_(dct), decoder_(0),
		 ibit_(-1), h261_rtp_bug_(0)
{}

H261Transcoder::~H261Transcoder()
{
	delete decoder_;
}

void H261Transcoder::configure()
{
	delete decoder_;
	if (ibit_ != 0) {
		if (dct_)
			decoder_ = new IntraP64DCTDecoder();
		else 
			decoder_ = new IntraP64Decoder();
	} else
		decoder_ = new FullP64Decoder();

	decoder_->mark(CR_SEND|CR_MOTION);
}

int H261Transcoder::recv(const rtphdr* rh, const u_char* bp, int cc, 
			     Transmitter& tx)
{		
	u_int v = ntohl(*(u_int*)(rh + 1));
	int sbit = v >> 29;
	int ebit = (v >> 26) & 7;
	int quant = (v >> 10) & 0x1f;
	int mvdh = (v >> 5) & 0x1f;
	int mvdv = v & 0x1f;
        int mba, gob;
	/*
	 * vic-2.7 swapped the GOB and MBA fields in the RTP packet header
	 * with respect to the spec.  To maintain backward compat, whenever
	 * we see an out of range gob, we change our assumption about the
	 * stream and continue.
	 */
	if (!h261_rtp_bug_) {
		mba = (v >> 15) & 0x1f;
		gob = (v >> 20) & 0xf;
		if (gob > 12) {
			h261_rtp_bug_ = 1;
			mba = (v >> 19) & 0x1f;
			gob = (v >> 15) & 0xf;
		}
	} else {
		mba = (v >> 19) & 0x1f;
		gob = (v >> 15) & 0xf;
		if (gob > 12) {
			h261_rtp_bug_ = 0;
			mba = (v >> 15) & 0x1f;
			gob = (v >> 20) & 0xf;
		}
	}

	int ibit = (v >> 25) & 1;
	if (ibit != ibit_) {
		ibit_ = ibit;
		configure();
	}

	bp += sizeof(h261hdr);
	cc -= sizeof(h261hdr);

	(void)decoder_->decode(bp, cc, sbit, ebit, mba, gob, quant, mvdh, 
			       mvdv);
	/*
	 * If the stream changes format, issue a resize.
	 */
	if (decoder_->width() != inw_ || decoder_->height() != inh_) {
		inw_ = decoder_->width();
		inh_ = decoder_->height();
		encoder_->framesize(inw_, inh_, outw_, outh_);
		crinit(outw_, outh_);
		decoder_->marks(crvec_);
	}

	if (!(ntohs(rh->rh_flags) & RTP_M))
		return (0);
	
	decoder_->sync();

	/* 
	 * Have a frame.
	 */
	if (tx.bps() == 0)
		return (1);

	double now = gettimeofday();
	if (fc_ <= now) {
		/* If we have fallen behind (>200ms), re-sync. */
		if (now - fc_ > 200000.) 	
			fc_ = now;
		VideoFrame vf(ntohl(rh->rh_ts), decoder_->frame(), outw_, 
			      outh_, 411, crvec_);
		int ecc = encoder_->consume(&vf, tx);
		age_blocks();
		double bits = 8 * double(ecc);
		fc_ += 1e6 * bits / tx.bps();
	}
	return (1);
}
